# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import dataclasses
import json  # type: ignore
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
import warnings

from google.api_core import gapic_v1, operations_v1, rest_helpers, rest_streaming
from google.api_core import exceptions as core_exceptions
from google.api_core import retry as retries
from google.auth import credentials as ga_credentials  # type: ignore
from google.auth.transport.requests import AuthorizedSession  # type: ignore
from google.cloud.location import locations_pb2  # type: ignore
from google.iam.v1 import iam_policy_pb2  # type: ignore
from google.iam.v1 import policy_pb2  # type: ignore
from google.longrunning import operations_pb2  # type: ignore
from google.protobuf import json_format
from requests import __version__ as requests_version

from google.cloud.functions_v2.types import functions

from .base import DEFAULT_CLIENT_INFO as BASE_DEFAULT_CLIENT_INFO
from .rest_base import _BaseFunctionServiceRestTransport

try:
    OptionalRetry = Union[retries.Retry, gapic_v1.method._MethodDefault, None]
except AttributeError:  # pragma: NO COVER
    OptionalRetry = Union[retries.Retry, object, None]  # type: ignore


DEFAULT_CLIENT_INFO = gapic_v1.client_info.ClientInfo(
    gapic_version=BASE_DEFAULT_CLIENT_INFO.gapic_version,
    grpc_version=None,
    rest_version=f"requests@{requests_version}",
)


class FunctionServiceRestInterceptor:
    """Interceptor for FunctionService.

    Interceptors are used to manipulate requests, request metadata, and responses
    in arbitrary ways.
    Example use cases include:
    * Logging
    * Verifying requests according to service or custom semantics
    * Stripping extraneous information from responses

    These use cases and more can be enabled by injecting an
    instance of a custom subclass when constructing the FunctionServiceRestTransport.

    .. code-block:: python
        class MyCustomFunctionServiceInterceptor(FunctionServiceRestInterceptor):
            def pre_create_function(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_create_function(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_delete_function(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_delete_function(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_generate_download_url(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_generate_download_url(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_generate_upload_url(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_generate_upload_url(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_get_function(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_get_function(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_list_functions(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_list_functions(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_list_runtimes(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_list_runtimes(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_update_function(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_update_function(self, response):
                logging.log(f"Received response: {response}")
                return response

        transport = FunctionServiceRestTransport(interceptor=MyCustomFunctionServiceInterceptor())
        client = FunctionServiceClient(transport=transport)


    """

    def pre_create_function(
        self,
        request: functions.CreateFunctionRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.CreateFunctionRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for create_function

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_create_function(
        self, response: operations_pb2.Operation
    ) -> operations_pb2.Operation:
        """Post-rpc interceptor for create_function

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_delete_function(
        self,
        request: functions.DeleteFunctionRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.DeleteFunctionRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for delete_function

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_delete_function(
        self, response: operations_pb2.Operation
    ) -> operations_pb2.Operation:
        """Post-rpc interceptor for delete_function

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_generate_download_url(
        self,
        request: functions.GenerateDownloadUrlRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.GenerateDownloadUrlRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for generate_download_url

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_generate_download_url(
        self, response: functions.GenerateDownloadUrlResponse
    ) -> functions.GenerateDownloadUrlResponse:
        """Post-rpc interceptor for generate_download_url

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_generate_upload_url(
        self,
        request: functions.GenerateUploadUrlRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.GenerateUploadUrlRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for generate_upload_url

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_generate_upload_url(
        self, response: functions.GenerateUploadUrlResponse
    ) -> functions.GenerateUploadUrlResponse:
        """Post-rpc interceptor for generate_upload_url

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_get_function(
        self, request: functions.GetFunctionRequest, metadata: Sequence[Tuple[str, str]]
    ) -> Tuple[functions.GetFunctionRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for get_function

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_get_function(self, response: functions.Function) -> functions.Function:
        """Post-rpc interceptor for get_function

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_list_functions(
        self,
        request: functions.ListFunctionsRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.ListFunctionsRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for list_functions

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_list_functions(
        self, response: functions.ListFunctionsResponse
    ) -> functions.ListFunctionsResponse:
        """Post-rpc interceptor for list_functions

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_list_runtimes(
        self,
        request: functions.ListRuntimesRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.ListRuntimesRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for list_runtimes

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_list_runtimes(
        self, response: functions.ListRuntimesResponse
    ) -> functions.ListRuntimesResponse:
        """Post-rpc interceptor for list_runtimes

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_update_function(
        self,
        request: functions.UpdateFunctionRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[functions.UpdateFunctionRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for update_function

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_update_function(
        self, response: operations_pb2.Operation
    ) -> operations_pb2.Operation:
        """Post-rpc interceptor for update_function

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_list_locations(
        self,
        request: locations_pb2.ListLocationsRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[locations_pb2.ListLocationsRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for list_locations

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_list_locations(
        self, response: locations_pb2.ListLocationsResponse
    ) -> locations_pb2.ListLocationsResponse:
        """Post-rpc interceptor for list_locations

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_get_iam_policy(
        self,
        request: iam_policy_pb2.GetIamPolicyRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[iam_policy_pb2.GetIamPolicyRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for get_iam_policy

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_get_iam_policy(self, response: policy_pb2.Policy) -> policy_pb2.Policy:
        """Post-rpc interceptor for get_iam_policy

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_set_iam_policy(
        self,
        request: iam_policy_pb2.SetIamPolicyRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[iam_policy_pb2.SetIamPolicyRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for set_iam_policy

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_set_iam_policy(self, response: policy_pb2.Policy) -> policy_pb2.Policy:
        """Post-rpc interceptor for set_iam_policy

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_test_iam_permissions(
        self,
        request: iam_policy_pb2.TestIamPermissionsRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[iam_policy_pb2.TestIamPermissionsRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for test_iam_permissions

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_test_iam_permissions(
        self, response: iam_policy_pb2.TestIamPermissionsResponse
    ) -> iam_policy_pb2.TestIamPermissionsResponse:
        """Post-rpc interceptor for test_iam_permissions

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_get_operation(
        self,
        request: operations_pb2.GetOperationRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[operations_pb2.GetOperationRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for get_operation

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_get_operation(
        self, response: operations_pb2.Operation
    ) -> operations_pb2.Operation:
        """Post-rpc interceptor for get_operation

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response

    def pre_list_operations(
        self,
        request: operations_pb2.ListOperationsRequest,
        metadata: Sequence[Tuple[str, str]],
    ) -> Tuple[operations_pb2.ListOperationsRequest, Sequence[Tuple[str, str]]]:
        """Pre-rpc interceptor for list_operations

        Override in a subclass to manipulate the request or metadata
        before they are sent to the FunctionService server.
        """
        return request, metadata

    def post_list_operations(
        self, response: operations_pb2.ListOperationsResponse
    ) -> operations_pb2.ListOperationsResponse:
        """Post-rpc interceptor for list_operations

        Override in a subclass to manipulate the response
        after it is returned by the FunctionService server but before
        it is returned to user code.
        """
        return response


@dataclasses.dataclass
class FunctionServiceRestStub:
    _session: AuthorizedSession
    _host: str
    _interceptor: FunctionServiceRestInterceptor


class FunctionServiceRestTransport(_BaseFunctionServiceRestTransport):
    """REST backend synchronous transport for FunctionService.

    Google Cloud Functions is used to deploy functions that are executed
    by Google in response to various events. Data connected with that
    event is passed to a function as the input data.

    A **function** is a resource which describes a function that should
    be executed and how it is triggered.

    This class defines the same methods as the primary client, so the
    primary client can load the underlying transport implementation
    and call it.

    It sends JSON representations of protocol buffers over HTTP/1.1
    """

    def __init__(
        self,
        *,
        host: str = "cloudfunctions.googleapis.com",
        credentials: Optional[ga_credentials.Credentials] = None,
        credentials_file: Optional[str] = None,
        scopes: Optional[Sequence[str]] = None,
        client_cert_source_for_mtls: Optional[Callable[[], Tuple[bytes, bytes]]] = None,
        quota_project_id: Optional[str] = None,
        client_info: gapic_v1.client_info.ClientInfo = DEFAULT_CLIENT_INFO,
        always_use_jwt_access: Optional[bool] = False,
        url_scheme: str = "https",
        interceptor: Optional[FunctionServiceRestInterceptor] = None,
        api_audience: Optional[str] = None,
    ) -> None:
        """Instantiate the transport.

        Args:
            host (Optional[str]):
                 The hostname to connect to (default: 'cloudfunctions.googleapis.com').
            credentials (Optional[google.auth.credentials.Credentials]): The
                authorization credentials to attach to requests. These
                credentials identify the application to the service; if none
                are specified, the client will attempt to ascertain the
                credentials from the environment.

            credentials_file (Optional[str]): A file with credentials that can
                be loaded with :func:`google.auth.load_credentials_from_file`.
                This argument is ignored if ``channel`` is provided.
            scopes (Optional(Sequence[str])): A list of scopes. This argument is
                ignored if ``channel`` is provided.
            client_cert_source_for_mtls (Callable[[], Tuple[bytes, bytes]]): Client
                certificate to configure mutual TLS HTTP channel. It is ignored
                if ``channel`` is provided.
            quota_project_id (Optional[str]): An optional project to use for billing
                and quota.
            client_info (google.api_core.gapic_v1.client_info.ClientInfo):
                The client info used to send a user-agent string along with
                API requests. If ``None``, then default info will be used.
                Generally, you only need to set this if you are developing
                your own client library.
            always_use_jwt_access (Optional[bool]): Whether self signed JWT should
                be used for service account credentials.
            url_scheme: the protocol scheme for the API endpoint.  Normally
                "https", but for testing or local servers,
                "http" can be specified.
        """
        # Run the base constructor
        # TODO(yon-mg): resolve other ctor params i.e. scopes, quota, etc.
        # TODO: When custom host (api_endpoint) is set, `scopes` must *also* be set on the
        # credentials object
        super().__init__(
            host=host,
            credentials=credentials,
            client_info=client_info,
            always_use_jwt_access=always_use_jwt_access,
            url_scheme=url_scheme,
            api_audience=api_audience,
        )
        self._session = AuthorizedSession(
            self._credentials, default_host=self.DEFAULT_HOST
        )
        self._operations_client: Optional[operations_v1.AbstractOperationsClient] = None
        if client_cert_source_for_mtls:
            self._session.configure_mtls_channel(client_cert_source_for_mtls)
        self._interceptor = interceptor or FunctionServiceRestInterceptor()
        self._prep_wrapped_messages(client_info)

    @property
    def operations_client(self) -> operations_v1.AbstractOperationsClient:
        """Create the client designed to process long-running operations.

        This property caches on the instance; repeated calls return the same
        client.
        """
        # Only create a new client if we do not already have one.
        if self._operations_client is None:
            http_options: Dict[str, List[Dict[str, str]]] = {
                "google.longrunning.Operations.GetOperation": [
                    {
                        "method": "get",
                        "uri": "/v2/{name=projects/*/locations/*/operations/*}",
                    },
                ],
                "google.longrunning.Operations.ListOperations": [
                    {
                        "method": "get",
                        "uri": "/v2/{name=projects/*/locations/*}/operations",
                    },
                ],
            }

            rest_transport = operations_v1.OperationsRestTransport(
                host=self._host,
                # use the credentials which are saved
                credentials=self._credentials,
                scopes=self._scopes,
                http_options=http_options,
                path_prefix="v2",
            )

            self._operations_client = operations_v1.AbstractOperationsClient(
                transport=rest_transport
            )

        # Return the client from cache.
        return self._operations_client

    class _CreateFunction(
        _BaseFunctionServiceRestTransport._BaseCreateFunction, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.CreateFunction")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
            )
            return response

        def __call__(
            self,
            request: functions.CreateFunctionRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> operations_pb2.Operation:
            r"""Call the create function method over HTTP.

            Args:
                request (~.functions.CreateFunctionRequest):
                    The request object. Request for the ``CreateFunction`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.operations_pb2.Operation:
                    This resource represents a
                long-running operation that is the
                result of a network API call.

            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseCreateFunction._get_http_options()
            )
            request, metadata = self._interceptor.pre_create_function(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseCreateFunction._get_transcoded_request(
                http_options, request
            )

            body = _BaseFunctionServiceRestTransport._BaseCreateFunction._get_request_body_json(
                transcoded_request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseCreateFunction._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._CreateFunction._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
                body,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = operations_pb2.Operation()
            json_format.Parse(response.content, resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_create_function(resp)
            return resp

    class _DeleteFunction(
        _BaseFunctionServiceRestTransport._BaseDeleteFunction, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.DeleteFunction")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: functions.DeleteFunctionRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> operations_pb2.Operation:
            r"""Call the delete function method over HTTP.

            Args:
                request (~.functions.DeleteFunctionRequest):
                    The request object. Request for the ``DeleteFunction`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.operations_pb2.Operation:
                    This resource represents a
                long-running operation that is the
                result of a network API call.

            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseDeleteFunction._get_http_options()
            )
            request, metadata = self._interceptor.pre_delete_function(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseDeleteFunction._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseDeleteFunction._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._DeleteFunction._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = operations_pb2.Operation()
            json_format.Parse(response.content, resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_delete_function(resp)
            return resp

    class _GenerateDownloadUrl(
        _BaseFunctionServiceRestTransport._BaseGenerateDownloadUrl,
        FunctionServiceRestStub,
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.GenerateDownloadUrl")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
            )
            return response

        def __call__(
            self,
            request: functions.GenerateDownloadUrlRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> functions.GenerateDownloadUrlResponse:
            r"""Call the generate download url method over HTTP.

            Args:
                request (~.functions.GenerateDownloadUrlRequest):
                    The request object. Request of ``GenerateDownloadUrl`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.functions.GenerateDownloadUrlResponse:
                    Response of ``GenerateDownloadUrl`` method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseGenerateDownloadUrl._get_http_options()
            )
            request, metadata = self._interceptor.pre_generate_download_url(
                request, metadata
            )
            transcoded_request = _BaseFunctionServiceRestTransport._BaseGenerateDownloadUrl._get_transcoded_request(
                http_options, request
            )

            body = _BaseFunctionServiceRestTransport._BaseGenerateDownloadUrl._get_request_body_json(
                transcoded_request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseGenerateDownloadUrl._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._GenerateDownloadUrl._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
                body,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = functions.GenerateDownloadUrlResponse()
            pb_resp = functions.GenerateDownloadUrlResponse.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_generate_download_url(resp)
            return resp

    class _GenerateUploadUrl(
        _BaseFunctionServiceRestTransport._BaseGenerateUploadUrl,
        FunctionServiceRestStub,
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.GenerateUploadUrl")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
            )
            return response

        def __call__(
            self,
            request: functions.GenerateUploadUrlRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> functions.GenerateUploadUrlResponse:
            r"""Call the generate upload url method over HTTP.

            Args:
                request (~.functions.GenerateUploadUrlRequest):
                    The request object. Request of ``GenerateSourceUploadUrl`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.functions.GenerateUploadUrlResponse:
                    Response of ``GenerateSourceUploadUrl`` method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseGenerateUploadUrl._get_http_options()
            )
            request, metadata = self._interceptor.pre_generate_upload_url(
                request, metadata
            )
            transcoded_request = _BaseFunctionServiceRestTransport._BaseGenerateUploadUrl._get_transcoded_request(
                http_options, request
            )

            body = _BaseFunctionServiceRestTransport._BaseGenerateUploadUrl._get_request_body_json(
                transcoded_request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseGenerateUploadUrl._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._GenerateUploadUrl._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
                body,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = functions.GenerateUploadUrlResponse()
            pb_resp = functions.GenerateUploadUrlResponse.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_generate_upload_url(resp)
            return resp

    class _GetFunction(
        _BaseFunctionServiceRestTransport._BaseGetFunction, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.GetFunction")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: functions.GetFunctionRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> functions.Function:
            r"""Call the get function method over HTTP.

            Args:
                request (~.functions.GetFunctionRequest):
                    The request object. Request for the ``GetFunction`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.functions.Function:
                    Describes a Cloud Function that
                contains user computation executed in
                response to an event. It encapsulates
                function and trigger configurations.

            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseGetFunction._get_http_options()
            )
            request, metadata = self._interceptor.pre_get_function(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseGetFunction._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseGetFunction._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._GetFunction._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = functions.Function()
            pb_resp = functions.Function.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_get_function(resp)
            return resp

    class _ListFunctions(
        _BaseFunctionServiceRestTransport._BaseListFunctions, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.ListFunctions")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: functions.ListFunctionsRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> functions.ListFunctionsResponse:
            r"""Call the list functions method over HTTP.

            Args:
                request (~.functions.ListFunctionsRequest):
                    The request object. Request for the ``ListFunctions`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.functions.ListFunctionsResponse:
                    Response for the ``ListFunctions`` method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseListFunctions._get_http_options()
            )
            request, metadata = self._interceptor.pre_list_functions(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseListFunctions._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseListFunctions._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._ListFunctions._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = functions.ListFunctionsResponse()
            pb_resp = functions.ListFunctionsResponse.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_list_functions(resp)
            return resp

    class _ListRuntimes(
        _BaseFunctionServiceRestTransport._BaseListRuntimes, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.ListRuntimes")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: functions.ListRuntimesRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> functions.ListRuntimesResponse:
            r"""Call the list runtimes method over HTTP.

            Args:
                request (~.functions.ListRuntimesRequest):
                    The request object. Request for the ``ListRuntimes`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.functions.ListRuntimesResponse:
                    Response for the ``ListRuntimes`` method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseListRuntimes._get_http_options()
            )
            request, metadata = self._interceptor.pre_list_runtimes(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseListRuntimes._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseListRuntimes._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._ListRuntimes._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = functions.ListRuntimesResponse()
            pb_resp = functions.ListRuntimesResponse.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_list_runtimes(resp)
            return resp

    class _UpdateFunction(
        _BaseFunctionServiceRestTransport._BaseUpdateFunction, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.UpdateFunction")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
            )
            return response

        def __call__(
            self,
            request: functions.UpdateFunctionRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> operations_pb2.Operation:
            r"""Call the update function method over HTTP.

            Args:
                request (~.functions.UpdateFunctionRequest):
                    The request object. Request for the ``UpdateFunction`` method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                ~.operations_pb2.Operation:
                    This resource represents a
                long-running operation that is the
                result of a network API call.

            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseUpdateFunction._get_http_options()
            )
            request, metadata = self._interceptor.pre_update_function(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseUpdateFunction._get_transcoded_request(
                http_options, request
            )

            body = _BaseFunctionServiceRestTransport._BaseUpdateFunction._get_request_body_json(
                transcoded_request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseUpdateFunction._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._UpdateFunction._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
                body,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = operations_pb2.Operation()
            json_format.Parse(response.content, resp, ignore_unknown_fields=True)
            resp = self._interceptor.post_update_function(resp)
            return resp

    @property
    def create_function(
        self,
    ) -> Callable[[functions.CreateFunctionRequest], operations_pb2.Operation]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._CreateFunction(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def delete_function(
        self,
    ) -> Callable[[functions.DeleteFunctionRequest], operations_pb2.Operation]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._DeleteFunction(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def generate_download_url(
        self,
    ) -> Callable[
        [functions.GenerateDownloadUrlRequest], functions.GenerateDownloadUrlResponse
    ]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._GenerateDownloadUrl(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def generate_upload_url(
        self,
    ) -> Callable[
        [functions.GenerateUploadUrlRequest], functions.GenerateUploadUrlResponse
    ]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._GenerateUploadUrl(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def get_function(
        self,
    ) -> Callable[[functions.GetFunctionRequest], functions.Function]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._GetFunction(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def list_functions(
        self,
    ) -> Callable[[functions.ListFunctionsRequest], functions.ListFunctionsResponse]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._ListFunctions(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def list_runtimes(
        self,
    ) -> Callable[[functions.ListRuntimesRequest], functions.ListRuntimesResponse]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._ListRuntimes(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def update_function(
        self,
    ) -> Callable[[functions.UpdateFunctionRequest], operations_pb2.Operation]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._UpdateFunction(self._session, self._host, self._interceptor)  # type: ignore

    @property
    def list_locations(self):
        return self._ListLocations(self._session, self._host, self._interceptor)  # type: ignore

    class _ListLocations(
        _BaseFunctionServiceRestTransport._BaseListLocations, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.ListLocations")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: locations_pb2.ListLocationsRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> locations_pb2.ListLocationsResponse:
            r"""Call the list locations method over HTTP.

            Args:
                request (locations_pb2.ListLocationsRequest):
                    The request object for ListLocations method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                locations_pb2.ListLocationsResponse: Response from ListLocations method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseListLocations._get_http_options()
            )
            request, metadata = self._interceptor.pre_list_locations(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseListLocations._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseListLocations._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._ListLocations._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            content = response.content.decode("utf-8")
            resp = locations_pb2.ListLocationsResponse()
            resp = json_format.Parse(content, resp)
            resp = self._interceptor.post_list_locations(resp)
            return resp

    @property
    def get_iam_policy(self):
        return self._GetIamPolicy(self._session, self._host, self._interceptor)  # type: ignore

    class _GetIamPolicy(
        _BaseFunctionServiceRestTransport._BaseGetIamPolicy, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.GetIamPolicy")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: iam_policy_pb2.GetIamPolicyRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> policy_pb2.Policy:
            r"""Call the get iam policy method over HTTP.

            Args:
                request (iam_policy_pb2.GetIamPolicyRequest):
                    The request object for GetIamPolicy method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                policy_pb2.Policy: Response from GetIamPolicy method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseGetIamPolicy._get_http_options()
            )
            request, metadata = self._interceptor.pre_get_iam_policy(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseGetIamPolicy._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseGetIamPolicy._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._GetIamPolicy._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            content = response.content.decode("utf-8")
            resp = policy_pb2.Policy()
            resp = json_format.Parse(content, resp)
            resp = self._interceptor.post_get_iam_policy(resp)
            return resp

    @property
    def set_iam_policy(self):
        return self._SetIamPolicy(self._session, self._host, self._interceptor)  # type: ignore

    class _SetIamPolicy(
        _BaseFunctionServiceRestTransport._BaseSetIamPolicy, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.SetIamPolicy")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
            )
            return response

        def __call__(
            self,
            request: iam_policy_pb2.SetIamPolicyRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> policy_pb2.Policy:
            r"""Call the set iam policy method over HTTP.

            Args:
                request (iam_policy_pb2.SetIamPolicyRequest):
                    The request object for SetIamPolicy method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                policy_pb2.Policy: Response from SetIamPolicy method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseSetIamPolicy._get_http_options()
            )
            request, metadata = self._interceptor.pre_set_iam_policy(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseSetIamPolicy._get_transcoded_request(
                http_options, request
            )

            body = _BaseFunctionServiceRestTransport._BaseSetIamPolicy._get_request_body_json(
                transcoded_request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseSetIamPolicy._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._SetIamPolicy._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
                body,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            content = response.content.decode("utf-8")
            resp = policy_pb2.Policy()
            resp = json_format.Parse(content, resp)
            resp = self._interceptor.post_set_iam_policy(resp)
            return resp

    @property
    def test_iam_permissions(self):
        return self._TestIamPermissions(self._session, self._host, self._interceptor)  # type: ignore

    class _TestIamPermissions(
        _BaseFunctionServiceRestTransport._BaseTestIamPermissions,
        FunctionServiceRestStub,
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.TestIamPermissions")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
            )
            return response

        def __call__(
            self,
            request: iam_policy_pb2.TestIamPermissionsRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> iam_policy_pb2.TestIamPermissionsResponse:
            r"""Call the test iam permissions method over HTTP.

            Args:
                request (iam_policy_pb2.TestIamPermissionsRequest):
                    The request object for TestIamPermissions method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                iam_policy_pb2.TestIamPermissionsResponse: Response from TestIamPermissions method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseTestIamPermissions._get_http_options()
            )
            request, metadata = self._interceptor.pre_test_iam_permissions(
                request, metadata
            )
            transcoded_request = _BaseFunctionServiceRestTransport._BaseTestIamPermissions._get_transcoded_request(
                http_options, request
            )

            body = _BaseFunctionServiceRestTransport._BaseTestIamPermissions._get_request_body_json(
                transcoded_request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseTestIamPermissions._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._TestIamPermissions._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
                body,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            content = response.content.decode("utf-8")
            resp = iam_policy_pb2.TestIamPermissionsResponse()
            resp = json_format.Parse(content, resp)
            resp = self._interceptor.post_test_iam_permissions(resp)
            return resp

    @property
    def get_operation(self):
        return self._GetOperation(self._session, self._host, self._interceptor)  # type: ignore

    class _GetOperation(
        _BaseFunctionServiceRestTransport._BaseGetOperation, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.GetOperation")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: operations_pb2.GetOperationRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> operations_pb2.Operation:
            r"""Call the get operation method over HTTP.

            Args:
                request (operations_pb2.GetOperationRequest):
                    The request object for GetOperation method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                operations_pb2.Operation: Response from GetOperation method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseGetOperation._get_http_options()
            )
            request, metadata = self._interceptor.pre_get_operation(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseGetOperation._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseGetOperation._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._GetOperation._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            content = response.content.decode("utf-8")
            resp = operations_pb2.Operation()
            resp = json_format.Parse(content, resp)
            resp = self._interceptor.post_get_operation(resp)
            return resp

    @property
    def list_operations(self):
        return self._ListOperations(self._session, self._host, self._interceptor)  # type: ignore

    class _ListOperations(
        _BaseFunctionServiceRestTransport._BaseListOperations, FunctionServiceRestStub
    ):
        def __hash__(self):
            return hash("FunctionServiceRestTransport.ListOperations")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None,
        ):
            uri = transcoded_request["uri"]
            method = transcoded_request["method"]
            headers = dict(metadata)
            headers["Content-Type"] = "application/json"
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
            )
            return response

        def __call__(
            self,
            request: operations_pb2.ListOperationsRequest,
            *,
            retry: OptionalRetry = gapic_v1.method.DEFAULT,
            timeout: Optional[float] = None,
            metadata: Sequence[Tuple[str, str]] = (),
        ) -> operations_pb2.ListOperationsResponse:
            r"""Call the list operations method over HTTP.

            Args:
                request (operations_pb2.ListOperationsRequest):
                    The request object for ListOperations method.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, str]]): Strings which should be
                    sent along with the request as metadata.

            Returns:
                operations_pb2.ListOperationsResponse: Response from ListOperations method.
            """

            http_options = (
                _BaseFunctionServiceRestTransport._BaseListOperations._get_http_options()
            )
            request, metadata = self._interceptor.pre_list_operations(request, metadata)
            transcoded_request = _BaseFunctionServiceRestTransport._BaseListOperations._get_transcoded_request(
                http_options, request
            )

            # Jsonify the query params
            query_params = _BaseFunctionServiceRestTransport._BaseListOperations._get_query_params_json(
                transcoded_request
            )

            # Send the request
            response = FunctionServiceRestTransport._ListOperations._get_response(
                self._host,
                metadata,
                query_params,
                self._session,
                timeout,
                transcoded_request,
            )

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            content = response.content.decode("utf-8")
            resp = operations_pb2.ListOperationsResponse()
            resp = json_format.Parse(content, resp)
            resp = self._interceptor.post_list_operations(resp)
            return resp

    @property
    def kind(self) -> str:
        return "rest"

    def close(self):
        self._session.close()


__all__ = ("FunctionServiceRestTransport",)
